# Copyright © 2020 Hashmap, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#  TODO:  add back in when artifactz is published to a repo
# from artifactz.artifact_registry import ArtifactRegistry
from providah.factories.package_factory import PackageFactory as pf

from servz.core.build_step import BuildStep


class PackagePublisher(BuildStep):

    def __init__(self, **kwargs):
        super().__init__()
        self._build_flows: list = list()

        artifactz_config = kwargs['artifactz']
        #  self.__artifact_registry: ArtifactRegistry = pf.create(key=artifactz_config['type'], configuration=artifactz_config['conf'])

    def build_part(self, **kwargs):

        raise NotImplementedError(f'Method not implemented for {type(self).__name__}.')

    def get_results(self):

        return {
            "package_publisher": self._build_flows
        }



# Copyright © 2020 Hashmap, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import yaml

from servz.core.pipline_composer.pipeline_composer import PipelineComposer
from servz.utils.exceptions.pipeline_read_error import PipelineReadError


class YAMLPipelineComposer(PipelineComposer):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.__root_path = kwargs.get('project_path')
        self.__workflows_path = os.path.join(kwargs.get('project_path'), 'workflows')
        self._error_messages = []

    def _build(self, flows: dict) -> list:
        """
        Build and return a dict containing the information about all pipelines that will need to be built. The contents
        of the specification will be validated.

        Args:
            flows: dictionary of all active flows

        Returns: flushed out pipeline flows as will be needed for building the containers

        """
        # Future Version will need to handle different schema versions via factory
        return [self.__build_flow(pipeline) for pipeline in flows['pipelines']]

    def __build_flow(self, pipeline: dict) -> None:

        # Read the pipeline yaml file
        pipeline_config_path = os.path.join(self.__workflows_path, 'pipelines', pipeline['name'] + '.yml')

        with open(pipeline_config_path, 'r') as fh:
            pipeline = yaml.safe_load(fh)

        self.__validate_pipeline(pipeline)

        if len(self._error_messages) > 0:
            self.__log_build_results()
            raise PipelineReadError("\n".join(self._error_messages))

        # get the workflow
        self._pipelines.append(pipeline)

    def __validate_pipeline(self, pipeline: dict) -> None:

        keys = [key.lower() for key in pipeline.keys()]
        required_keys = {'project_root', 'workflow', 'version', 'stages'}
        # Stages has root and workflow
        if not required_keys.intersection(keys):
            self._generate_validation_message(f'At least one of {required_keys} is missing from configuration.')

        workflow: dict = pipeline.get('workflow')
        # Workflow is list
        if not isinstance(workflow, list):
            self._generate_validation_message(f'"workflow" must be a list.')

        # Workflow components are dicts
        if any([1 for node in workflow if not isinstance(node, dict)]):
            self._generate_validation_message(f'All entries in the workflow must be a dictionary.')

        # Workflow component path and files exist and is a python file
        _ = [self.__validate_nodes(node=node) for node in workflow]

        # Get list of all dependencies
        dependency_lists = [node.get('dependencies') for node in workflow if 'dependencies' in node.keys()]
        dependencies = []
        _ = [dependencies.extend(dependency) for dependency in dependency_lists]
        dependencies = list(set(dependencies))

        _ = [self.__validate_dependencies(all_dependencies=dependencies, node=node) for node in workflow]

    def __log_build_results(self) -> None:
        self._logger.error("\n".join(self._error_messages))

    def __validate_nodes(self, node: dict) -> None:

        keys = [key.lower() for key in node.keys()]
        if 'stage_name' not in keys:
            self._generate_validation_message("in workflow not 'stage_name' found.")

        if 'fit' not in keys:
            self._generate_validation_message("in workflow not 'fit' found.")

        if 'type' not in keys:
            self._generate_validation_message("in workflow not 'type' found.")

        entrypoint = os.path.join(self.__root_path, node.get('fit'))
        if not os.path.exists(entrypoint):
            self._generate_validation_message(f'fit {entrypoint} does not exist.')

        if not os.path.isfile(entrypoint):
            self._generate_validation_message(f'fit {entrypoint} is not a file')

    def _generate_validation_message(self, message: str) -> None:
        return self._error_messages.append(f'In {__name__}: {message}')

    def __validate_dependencies(self, all_dependencies: list, node: dict) -> None:

        keys = [key.lower() for key in node.keys()]

        if 'dependencies' not in keys:
            return

        dependencies = node.get('dependencies')
        if not isinstance(dependencies, list):
            self._generate_validation_message(f'dependencies is not a list.')

        if any([1 for dependency in dependencies if dependency not in all_dependencies]):
            self._generate_validation_message(f'dependency was not tracked in ')



# Copyright © 2020 Hashmap, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import traceback
from jinja2 import Template
import uuid
from servz.core.endpoint_appender.endpoint_appender import EndpointAppender
from servz.utils.parsers.project_config import ProjectConfig


class FlaskAppender(EndpointAppender):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        print(kwargs,"---")
        self.__root_path: str = os.path.abspath(kwargs.get('project_path'))
        self.__prediction_main = ""
        self.__predictor = ''

    def build_runner_script(self, pipe):

        try:

            print("   -->>>>    build_runner_script wrf flask  ", pipe)
            #  TODO:  template this out later
            runner = os.path.abspath(os.path.join(self.__root_path, 'server_' + uuid.uuid4().hex) + '.sh')
            requirements = 'pip install -r requirements_txt'
            executable_app = f'python app.py'
            # Save runner file to disk

            with open(runner, 'w') as fh:
                fh.write(requirements+"\n")
                fh.write(executable_app)
            fh.close()
        except:
            self.__handle_exception(f'The following error has occurred{traceback.format_exc()}')

    def build_artifact_bundle(self, pipe):

        self.__prediction_main = pipe.get('predict')
        path = str(self.__prediction_main)
        path_list = ".".join(path.split('/'))
        self.__predictor = path_list.replace('.py', '')

    def build_endpoint(self, pipe):

        # read in template file
        flask_code_file = os.path.join(ProjectConfig.package_root(), 'server_templates', 'flask', 'app')
        with open(flask_code_file, 'r') as fh:
            code = fh.read()

        # apply parameters
        flask_code = Template(code, autoescape=True)
        endpoint = flask_code.render(predict=self.__predictor)

        # save out to local path
        server = os.path.abspath(os.path.join(self.__root_path, 'app.py'))
        print(server)
        with open(server, 'w') as fh:
            fh.write(endpoint)



# Copyright © 2020 Hashmap, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from servz.core.build_step import BuildStep
from servz.utils.exceptions.flow_build_error import FlowBuildError


class PipelineComposer(BuildStep):

    def __init__(self, **kwargs):
        super().__init__()
        self._pipelines: list = []

    def build_part(self, **kwargs):
        _res = self._build(flows=kwargs.get("workflows"))

        if not _res:
            raise FlowBuildError()

    def get_results(self):

        result = dict()
        if self._pipelines:
            result = {
                "pipelines": self._pipelines
            }

        return result

    def _build(self, flows: dict) -> bool:
        raise NotImplementedError(f'Method not implemented for {type(self).__name__}.')



# Copyright © 2020 Hashmap, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import unittest
from servz.core.director import Director
import pathlib
import os


class TestDirector(unittest.TestCase):
    """
    Test Builder Class
    """

    def setUp(self) -> None:
        self.__path = pathlib.Path(__file__).parent.absolute()
        self._builder = Director(path=os.path.join(self.__path, 'ml_qookeys'))

    def tearDown(self) -> None:
        del self._builder

    def test_build_and_run(self):
        with self.assertRaises(Exception):
            self._builder.build_and_run()



# Copyright © 2020 Hashmap, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging


class BuildStep:
    def __init__(self):
        self._logger = logging.getLogger(__name__)
        self._logger.setLevel(logging.INFO)

    def _error_handler(self, exception_message: str) -> None:
        error_message = f"Error occurred in Source: {exception_message}"
        self._logger.exception(error_message)



# Copyright © 2020 Hashmap, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import yaml
from providah.factories.package_factory import PackageFactory as pf
from servz.core.build_pipeline import BuildPipeline
from servz.utils.parsers.project_config import ProjectConfig


class Builder:

    def __init__(self, **kwargs):

        self.__path = kwargs.get("project_path")
        self.__build_pipeline = BuildPipeline

    @property
    def pipeline(self):
        return self.__build_pipeline

    def build(self):
        # Configuration is lazy
        with open(ProjectConfig.config_path(), 'r') as stream:
            configuration = yaml.safe_load(stream)

        if self.__path:
            configuration['workflow_validation']['conf']['project_path'] = self.__path
            configuration['pipeline_composition']['conf']['project_path'] = self.__path
            configuration['orchestration_artifact_builder']['conf']['project_path'] = self.__path
            configuration['endpoint_appender']['conf']['project_path'] = self.__path
            configuration['packager']['conf']['project_path'] = self.__path
            configuration['package_publisher']['conf']['project_path'] = self.__path

        self.__build_pipeline.workflow_validator = pf.create(key=configuration['workflow_validation']['type'],
                                                         configuration=configuration['workflow_validation']['conf'])

        self.__build_pipeline.pipeline_composer = pf.create(key=configuration['pipeline_composition']['type'],
                                                            configuration=configuration['pipeline_composition']['conf'])

        self.__build_pipeline.deployment_artifact_builder = pf.create(key=configuration['orchestration_artifact_builder']['type'],
                                                                      configuration=configuration['orchestration_artifact_builder']['conf'])

        self.__build_pipeline.endpoint_appender = pf.create(key=configuration['endpoint_appender']['type'],
                                                            configuration=configuration['endpoint_appender']['conf'])

        self.__build_pipeline.packager = pf.create(key=configuration['packager']['type'],
                                                   configuration=configuration['packager']['conf'])

        self.__build_pipeline.package_publisher = pf.create(key=configuration['package_publisher']['type'],
                                                            configuration=configuration['package_publisher']['conf'])




# Copyright © 2020 Hashmap, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os

import traceback

from servz.core.packager.packager import Packager
from servz.utils.exceptions.packager_error import PackagerError
from servz.utils.parsers.config_parser import ConfigParser
from servz.utils.parsers.cmd_config_parser import CmdConfigParser


class PrefectFilePackager(Packager):

    # ---------------------------------------- #
    # ------------- Construction ------------- #
    # ---------------------------------------- #

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.__root_path: str = os.path.abspath(kwargs.get('project_path'))
        self.__workflows_path: str = os.path.join(kwargs.get('project_path'), 'workflows')

        self._workflows: list = list()
        self._pacified_tasks: list = list()
        self._file_list = []
        config_parser = ConfigParser()
        config_parser.parse()
        __cmd_config_obj = CmdConfigParser()
        __cmd_config_obj.parse()
        self.__cmd_config = __cmd_config_obj.configuration()

    def build_part(self, **kwargs):

        try:
            print("____-->>   build_part  prefect pkg ", kwargs)
            if 'pipelines' not in kwargs.keys():
                error_message = f"In {__name__}.build_part 'pipelines' was not found in build_part. " \
                                f"The parameters passed were: {' - '.join(kwargs.keys())}."
                raise ValueError(error_message)

            self._build(pipelines=kwargs.get("pipelines"))

        except:
            self.__handle_exception(f'The following error has occurred{traceback.format_exc()}')

    def _build(self, pipelines: list) -> None:
        self._workflows = []
        for pipeline in pipelines:
            updated_pipes: list = list()
            while len(pipeline['workflow']) > 0:
                for pipe in pipeline['workflow']:
                    _res = self.__build_file(pipe=pipe)
                    if _res['successful']:
                        # Update dependencies pacified
                        self._pacified_tasks.append(pipe['stage_name'])
                        # Make a copy of the new pipes
                        updated_pipes.append(pipe)
                        # Remove old pipe
                        pipeline['workflow'].remove(pipe)

            self._build_flows.append(
                {
                    'name': pipeline['name'],
                    'workflow': self._file_list
                }
            )

    def __build_file(self, pipe: dict) -> dict:
        if pipe.get("cmd_type"):
            pipe.update(self.__cmd_config[pipe["cmd_type"]])

        if os.path.exists(self.__root_path):
            for root, dirs, files in os.walk(self.__root_path):
                for file in files:
                    selected_file = os.path.join(root, file)
                    self._file_list.append(selected_file)

            self._logger.info('file built.')
            try:

                return dict(
                    successful=True

                )

            except:
                self.__handle_exception(f'The following error has occurred{traceback.format_exc()}')
        else:
            self.__handle_exception(f"file {self.__root_path} do not exist.")

    def __handle_exception(self, message: str) -> None:
        message = f'In {__name__}: {message}'
        self._logger.error(message)
        raise (PackagerError(message))

    def __are_dependencies_pacified(self, dependencies: list) -> bool:
        return len([1 for dependency in dependencies if dependency in self._pacified_tasks]) == len(dependencies)



# Copyright © 2020 Hashmap, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import NamedTuple
from servz.core.workflow_validator.workflow_validator import WorkflowValidator
from servz.core.pipline_composer.pipeline_composer import PipelineComposer
from servz.core.orchestration_artifact_builder.orchestration_artifact_builder import OrchestrationArtifactBuilder
from servz.core.endpoint_appender.endpoint_appender import EndpointAppender
from servz.core.packager.packager import Packager
from servz.core.package_publisher.package_publisher import PackagePublisher


class BuildPipeline(NamedTuple):
    workflow_validator: WorkflowValidator
    pipeline_composer: PipelineComposer
    deployment_artifact_builder: OrchestrationArtifactBuilder
    endpoint_appender: EndpointAppender
    packager: Packager
    package_publisher: PackagePublisher



# Modifications © 2020 Hashmap, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os

import yaml


class Parser:

    def __init__(self):
        self._path: str = ""
        self._config: dict = {}

    def configuration(self):
        return self._config

    def parse(self):
        with open(self._path, 'r') as file_stream:
            self._config = yaml.safe_load(file_stream)

        return self



# Copyright © 2020 Hashmap, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
from pathlib import Path

from servz.core.director import Director

testing_path = '../tests/ml_qookeys'

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Process the build and deploy an ML training process.')

    parser.add_argument('--project_path', '-p',
                        default=testing_path,
                        help='Path noting the root of the application.')

    args = parser.parse_args()

    args.project_path = str(Path(args.project_path).resolve())

    director = Director(project_path=args.project_path)

    director.build_and_run()



# Copyright © 2020 Hashmap, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import yaml

from servz.core.workflow_validator.workflow_validator import WorkflowValidator


class YAMLValidator(WorkflowValidator):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.__full_flow_path = os.path.join(kwargs.get('project_path'), 'workflows', 'flows.yml')

    def _validate(self) -> bool:
        """
        Read in configuration information (for flow reading), read the flow configuration and validate all contents.

        Returns: dictionary with the flows' pipelines

        Raises:
            FileNotFoundError -> flows.yml file was not found
            ValueError -> The contents of the flows.yml file were not correct
        """

        self._workflows = dict()

        if not self.__check_if_flow_is_present():
            raise FileNotFoundError('flows.yml not found')

        self.__read_flows()

        errors = self.__get_flow_validation_errors()
        if len(errors) > 0:
            raise ValueError('\n'.join(errors))

        self.__remove_inactive_pipelines()

        return True

    def __read_flows(self) -> None:
        """
        Read flows configuration file.

        """

        with open(self.__full_flow_path, 'r') as flows_stream:
            self._workflows = yaml.safe_load(flows_stream)

    def __check_if_flow_is_present(self) -> bool:
        """
        Check and see if 'flows' file is present.

        Returns: True if file is present, false otherwise.

        """

        return os.path.exists(self.__full_flow_path) and os.path.isfile(self.__full_flow_path)

    def __get_flow_validation_errors(self) -> list:
        """
        Validate the contents of the flows.yml configuration file.

        Returns: False if contents are not: a dictionary, do not contain version or pipelines and
        pipelines is not a list, True otherwise.

        """

        errors = []
        if not isinstance(self._workflows, dict):
            errors.append(f'Content of {self.__full_flow_path} is not a dictionary.')

        required_keys = {'version', 'pipelines'}
        if not required_keys.intersection(set(self._workflows.keys())):
            errors.append(f'At least one of the required keys - {required_keys} - was not present.')

        if not isinstance(self._workflows['pipelines'], list):
            errors.append(f'In {self.__full_flow_path}, pipelines is not a list.')

        return errors

    def __remove_inactive_pipelines(self) -> None:
        """
        Remove pipelines that are not active and have been labeled as such.

        """

        # Capture the pipelines that are active by removing those that are not active.
        pipelines = [flow for flow in self._workflows['pipelines'] if flow['active'] is not False]

        # Update the pipelines with the current ones.
        self._workflows['pipelines'] = pipelines



# Modifications © 2020 Hashmap, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging


class MLError(Exception):
    def __init__(self, *args):
        if args:
            self.message = args[0]
        else:
            self.message = None

        self._logger = logging.getLogger(__name__)
        self._logger.setLevel(logging.INFO)

    def __str__(self):
        if self.message:
            msg = f'{type(self).__name__}, {self.message}'
        else:
            msg = f'{type(self).__name__} has been raised.'
        self._logger.exception(msg)
        return msg



# Copyright © 2020 Hashmap, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from servz.core.build_step import BuildStep
from servz.utils.exceptions.orchestration_artifact_build_error import OrchestrationArtifactBuildError


class OrchestrationArtifactBuilder(BuildStep):

    def __init__(self, **kwargs):
        super().__init__()
        self._workflows: list = list()

    def build_part(self, **kwargs):
        print("--->>>  build part orches ", kwargs)
        _res = self._build(pipelines=kwargs.get('pipelines'))

        if not _res:
            raise OrchestrationArtifactBuildError()
        return True

    def get_results(self):

        result = {
            "artifact": self._workflows
        }

        return result

    def _build(self, pipelines: list) -> bool:
        print("--->>>  build builder orches ", pipelines)
        results = [self._compile_workflow(pipe) for pipe in pipelines]

        self._workflows = pipelines
        return all(wf['success'] for wf in results)

    def _compile_workflow(self, pipe):
        raise NotImplementedError(f'Method not implemented for {type(self).__name__}.')

    @classmethod
    def _generate_dependencies_string(cls, dependencies: list) -> str:
        if not dependencies:
            return ''

        return ','.join(dependencies)



# Copyright © 2020 Hashmap, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import uuid
from jinja2 import Template
from servz.core.orchestration_artifact_builder.orchestration_artifact_builder import OrchestrationArtifactBuilder
from servz.utils.parsers.project_config import ProjectConfig


class PrefectArtifactBuilder(OrchestrationArtifactBuilder):

    def __init__(self, **kwargs):

        super().__init__(**kwargs)
        print("*** Load Prefect Templates ****************")
        print(f"package root: {ProjectConfig.package_root()}")
        self.__manifest_templates = {
            'prefect_flow_template': os.path.join(ProjectConfig.package_root(), 'templates/prefect_templates',
                                                'prefect_flow_template'),
            'prefect_class_template': os.path.join(ProjectConfig.package_root(), 'templates/prefect_templates',
                                                  'prefect_class_template'),
            'prefect_edge_template': os.path.join(ProjectConfig.package_root(), 'templates/prefect_templates',
                                                  'prefect_edge_template'),
            'prefect_instance_template': os.path.join(ProjectConfig.package_root(), 'templates/prefect_templates',
                                                  'prefect_instance_template')

        }

        self.__namespace = kwargs.get('namespace')

    def _compile_workflow(self, pipe: dict) -> dict:
        print("   -->>>>    compile wrf prefect  ", pipe)
        classes: list = list()
        instances: list = list()
        edges: list = list()
        for task in pipe['workflow']:
            classes.append(self.__append_class(task))
            instances.append(self.__append_instance(task))
            edges.append(self.__append_edge(task))

        success = self.__build_prefect_flow(pipe=pipe, classes=classes, instances=instances, edges=edges)

        return {
            'success': success,
        }

    def __append_class(self, task):
        print("----- >  >  append class ", task )
        with open(self.__manifest_templates['prefect_class_template'], 'r') as f:
            class_temp = f.read()

        script = task.get("predict").replace("\\", "\\\\")

        class_template = Template(class_temp, autoescape=True)
        task_str = class_template.render(
            # TODO: replace this with camel case at some point
            class_name=task.get("stage_name").lower().replace(" ", "_"),
            command=f'{task.get("type")} {script.lower()}'
        )

        return task_str

    def __append_instance(self, task):
        with open(self.__manifest_templates['prefect_instance_template'], 'r') as f:
            instance_temp = f.read()

        node = task.get("stage_name").lower().replace(" ", "_")

        instance_template = Template(instance_temp, autoescape=True)
        task_str = instance_template.render(
            node=node,
            # TODO: replace this with camel case at some point
            instance=node.title()
        )

        return task_str

    def __append_edge(self, task):
        with open(self.__manifest_templates['prefect_edge_template'], 'r') as f:
            edge_temp = f.read()

        edge_str = ""
        deps = task.get('dependencies')
        if len(deps) == 0:
            return edge_str
        b = task["stage_name"].lower().replace(" ", "_")

        for dep in deps:
            edge_template = Template(edge_temp, autoescape=True)
            temp_str = edge_template.render(
                a=dep.lower().replace(" ", "_"),
                b=b
            )
            edge_str = f'{edge_str}\n{temp_str}'

        return edge_str

    def __build_prefect_flow(self, pipe: dict, classes: list, instances: list, edges: list):
        pipeline_name = pipe['name'].lower().replace(" ", "_")

        with open(self.__manifest_templates['prefect_flow_template'], 'r') as f:
            dag_temp = f.read()

        dag_template = Template(dag_temp, autoescape=True)

        airflow_str = dag_template.render(
            classes=''.join(classes),
            instances=''.join(instances),
            edges=''.join(edges),
        )

        pipeline_name = pipeline_name + '_prefect_' + uuid.uuid4().hex + '.py'
        # TODO This shouldn't be hard-coded and should be by 'run'
        with open(pipeline_name, 'w') as stream:
            stream.write(airflow_str)

        pipe['artifact_name'] = pipeline_name

        return True



# Copyright © 2020 Hashmap, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import platform
import os
from pathlib import Path


class ProjectConfig:
    @classmethod
    def home(cls):
        if not os.getenv('SERVZ'):
            if platform.system().lower() != 'windows':
                os.environ['SERVZ'] = os.getenv('HOME')
            else:
                os.environ['SERVZ'] = os.getenv('USERPROFILE')
        return os.getenv('SERVZ')

    @classmethod
    def env_var(cls):
        return os.getenv('SERVZ')

    @classmethod
    def config_path(cls):
        return os.path.join(cls.home(), f'.{__name__.split(".")[0]}/configuration.yml')

    @classmethod
    def env_path(cls):
        return os.path.join(cls.home(), f'.{__name__.split(".")[0]}/command_config.yml')

    @classmethod
    def package_root(cls):
        return Path(__file__).parent.parent.parent



# Copyright © 2020 Hashmap, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="servz",
    version="0.0.0.36",
    author="Hashmap, and NTT DATA Services Company",
    author_email="accelerators@hashmapinc.com",
    description="DO NOT USE - This is a sample program",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/hashmapinc/servz",
    packages=setuptools.find_packages(),
    package_data={
        "servz": ["templates/prefect_templates/*", "templates/server_templates/flask/*"]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
    ],
    python_requires='>=3.6',
)



# Copyright © 2020 Hashmap, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import traceback
from servz.core.builder import Builder
from servz.utils.exceptions.container_build_error import ContainerBuildError
from servz.utils.exceptions.flow_build_error import FlowBuildError
from servz.utils.exceptions.flow_validation_error import FlowValidationError


class Director:

    def __init__(self, **kwargs):
        self._logger = logging.getLogger(__name__)
        self._logger.setLevel(logging.INFO)
        # TODO Validate Builder steps are present

        # TODO This should be hidden in some way - maybe
        self._builder = Builder(**kwargs)

    def build_and_run(self) -> dict:
        """

        Returns:
            run summary as dict

        Raises:
            FlowValidationError

        """
        # ---------------------------- #
        # -------- Initialize -------- #
        # ---------------------------- #

        self._builder.build()

        run_summary = dict()

        try:

            self._builder.pipeline.workflow_validator.validate()

            self._builder.pipeline.pipeline_composer.build_part(**self._builder.pipeline.workflow_validator.get_results())

            self._builder.pipeline.deployment_artifact_builder.build_part(**self._builder.pipeline.pipeline_composer.get_results())

            self._builder.pipeline.endpoint_appender.build_part(**self._builder.pipeline.pipeline_composer.get_results())

            self._builder.pipeline.packager.build_part(**self._builder.pipeline.pipeline_composer.get_results())

            self._builder.pipeline.package_publisher.build_part(**self._builder.pipeline.pipeline_composer.get_results())

        except FlowValidationError:
            raise
        except FlowBuildError:
            raise
        except ContainerBuildError:
            raise
        except Exception:
            error_message = traceback.format_exc()
            self._error_handler(error_message)
            raise

        finally:
            self.__log_build_results(log_info=run_summary)

        # TODO: Summary needs to be built
        return run_summary

    def __log_build_results(self, log_info: dict) -> bool:
        pass

    def _error_handler(self, exception_message: str) -> None:
        error_message = f"Error occured in Source: {exception_message}"
        self._logger.exception(error_message)


# Copyright © 2020 Hashmap, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from servz.core.package_publisher.package_publisher import PackagePublisher


class ZipfilePublisher(PackagePublisher):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def build_part(self, **kwargs):

        result = self.__artifact_registry.register('test', artifact={
            'filename': kwargs['zip_file']
        })

        self._build_flows.append({'register_result': result})

    def get_results(self):

        return {
            "package_publisher": self._build_flows
        }
